/*
 * Some variables to facilitate passing arguments from the main process to the renderer process. 
 */

const WebMenuEvent = require('../web/web-menu-event');
const WebRequester = require('../web/web-requester');
const WebReceiver = require('../web/web-receiver');
const ElectronMenuEvent = require('../electron/electron-menu-event');
const ElectronRequester = require('../electron/electron-requester');
const ElectronReceiver = require('../electron/electron-receiver');
const Constants = require('../utils/constants');

const wss = (window.process && window.process.versions['electron']) ? null : new WebSocket('ws://' + window.location.hostname + ':8080');
const menuEvent = (window.process && window.process.versions['electron']) ? new ElectronMenuEvent() : new WebMenuEvent();
const requester = (window.process && window.process.versions['electron']) ? new ElectronRequester() : new WebRequester(wss);
const receiver = (window.process && window.process.versions['electron']) ? new ElectronReceiver() : new WebReceiver(wss);

const displayMode = {
    value: 'free',
    vr: false,
    spy: false,
    view: Constants.DisplayMode.HKL
};

menuEvent.on('change-mode', (event, value) => {
    displayMode.value = value;
});

module.exports = { displayMode, menuEvent, requester, receiver };