const state = require('../app-state');
const PubSub = require('pubsub-js');

function synchronize(data) {

    //console.log('synchronize ' + JSON.stringify(data));

    state.mode.update(data.mode);
    state.expdata.update(data.expdata);
    state.grid.update(data.grid);
    state.plane.update(data.plane);
    state.peaks.update(data.peaks);
    let updateSample = state.sample.update(data.sample);
    let updateAtoms = state.atoms.update(data.atoms);

    // The update is only done if it is not paused.
    state.scattering.update(data.scattering);
    
    let updateTrajectories = state.trajectory.update(data.trajectory);
    state.trajectoryData.update(data.trajectoryData);

    state.matrix.update(data.matrix);

    state.config.configurations = {};

    PubSub.publish('change-scattering', 0);
    PubSub.publish('change-plane', 0);

    PubSub.publish('change-sample', 0);

    PubSub.publish('change-peaks', 0);
    PubSub.publish('change-grid', 0);
    PubSub.publish('change-expdata', 0);

    PubSub.publish('change-trajectory', 0);
    PubSub.publish('update-trajectory-data', 0);
    
    PubSub.publish('change-atoms-conf', 0);
    PubSub.publish('update-atom-controller', 0);
}

function updateTrajectoryNames(trajectoriesModel) {
    
    let trajectories = state.trajectory.trajectories;

    for (let id in trajectoriesModel) {
        let traj = trajectoriesModel[id];

        if (id in trajectories) {
            let currentTraj = trajectories[id];

            if (currentTraj.name !== traj.name) {
                currentTraj.name = traj.name;
                PubSub.publish('change-trajectory-name', {id, name: traj.name});
            }
        }
    }
}

function synchronizeSpy(data) {
    
    //console.log('synchronize spy ' + JSON.stringify(data));
    
    state.mode.update(data.mode);
    //state.expdata.update(data.expdata);
    let updateGrid = state.grid.update(data.grid);
    let updatePlane = state.plane.update(data.plane);
    let updatePeaks = state.peaks.update(data.peaks);
    let updateSample = state.sample.update(data.sample);
    let updateAtoms = state.atoms.update(data.atoms);

    // The update is only done if it is not paused.
    state.scattering.update(data.scattering);
    
    updateTrajectoryNames(data.trajectory);
    let updateTrajectories = state.trajectory.update(data.trajectory);
    state.trajectoryData.update(data.trajectoryData);

    let updateMatrix = state.matrix.update(data.matrix);

    state.config.configurations = {};

    PubSub.publish('change-scattering', 0);

    if (updatePlane || updateMatrix) {
        PubSub.publish('change-plane', 0);
    }

    if (updateSample) {
        PubSub.publish('change-sample', 0);
        PubSub.publish('redraw-unitcell', { "spacegroup": state.sample.sg });
    }

    if (updatePeaks) {
        PubSub.publish('change-peaks', 0);
    }

    if (updateGrid) {
        PubSub.publish('change-grid', 0);
    }
//    PubSub.publish('change-expdata', 0);

    if (updateTrajectories) {
        PubSub.publish('change-trajectory', 0);
    }
    
    PubSub.publish('update-current-trajectory-data', 0);
    
    if (updateAtoms) {
        PubSub.publish('change-atoms-conf', 0);
        PubSub.publish('update-atom-controller', 0);
    }
}

function synchronizeScan(data) {

    state.scattering.update(data.scattering);
    state.trajectory.update(data.trajectory);

    PubSub.publish('change-scattering', 0);
    PubSub.publish('change-trajectory', 0);
}

module.exports = { synchronize, synchronizeScan, synchronizeSpy };