const math = require("mathjs");

// ----------------------------------------------------------------------------
// helper functions

const colorSheme = buildRainbowScale();

/**
 * map the given S(q,w) value to an RGB color triplet
 */
function mapSToColor(S, minS, maxS)
{
    const range = maxS - minS;
    const t = (S-minS) / range;
    let val = math.floor(t*255);

    if(val > 255) {
        val = 255;
    }

    let r = colorSheme.r[val];
    let g = colorSheme.g[val];
    let b = colorSheme.b[val];
    
    return { "r": r, "g": g, "b": b };
}

function buildRainbowScale() {
    let base_red   = [  0,   0,   0, 153, 255, 255, 255 ];
    let base_green = [  0,   0, 255, 255, 255, 153,   0 ];
    let base_blue  = [ 77, 204, 255,  77,   0,   0,   0 ];
    return interpolateColorScale(base_red, base_green, base_blue);
}

function interpolateColorScale(base_red, base_green, base_blue) {

    let red = new Array(256);
    let green = new Array(256);
    let blue = new Array(256);

    let n_ranges = base_red.length - 1;
    let out_length = red.length;
    let range_size = (out_length - 0.999) / n_ranges;

    let range = 0;
    let step_base = 0;

    for (let i = 0; i < out_length; i++ ) {
        if (i > (step_base + range_size)) {
            range = range + 1;
            step_base = step_base + range_size;
        }
        let t = (i - step_base) / range_size;

        red[i] = (((1.0-t) * base_red[range] + t * base_red[range+1]) || 0);
        green[i] = (((1.0-t) * base_green[range] + t * base_green[range+1]) || 0);
        blue[i] = (((1.0-t) * base_blue[range] + t * base_blue[range+1]) || 0);
    }

    return { r: red, g: green, b: blue };
}

// ----------------------------------------------------------------------------
/**
 * exports
 */
module.exports = {
    // helper functions
    "mapColor" : mapSToColor
};
// ----------------------------------------------------------------------------
