const viewManager = require('../../utils/view-manager');
const { scattering, matrix, mode } = require('../../app-state');
const { calculateKiKf } = require('../../utils/calculations');
const THREE = require('three');

class ScatteringPanelView {

    constructor() {

        let html = `
        <div class='container' style='width:100%'>
            <div class='row'>
                <label class='col-sm-2' style='color:rgb(255,0,0)' id='kiLabel'>|Ki|</label>
                <input type='text' class='col-sm-3' id='ki'/>
                <label class='col-sm-2'>&#8491;<sup>-1</sup></label>
                <div class='btn-group-toggle col-sm-4'>
                    <label class='btn btn-sm btn-outline-info'>
                        <input id='fixedKi' type='checkbox' data-toggle='ki' autocomplete='off'> Fixed
                    </label>
                </div>
            </div>
            <div class='row'>
                <label class='col-sm-2' style='color:rgb(0,128,0)' id='kfLabel'>|Kf|</label>
                <input type='text' class='col-sm-3' id='kf'/>
                <label class='col-sm-2'>&#8491;<sup>-1</sup></label>
                <div class='btn-group-toggle col-sm-4'>
                    <label class='btn btn-sm btn-outline-info'>
                        <input id='fixedKf' type='checkbox' data-toggle='kf' autocomplete='off'> Fixed
                    </label>
                </div>
            </div>
            <div class='row'>
                <label class='col-sm-2' style='color:rgb(0,0,255)' id='qLabel'>|Q|</label>
                <input type='text' class='col-sm-3' id='qNorm'/>
                <label class='col-sm-2'>&#8491;<sup>-1</sup></label>
                <div class='btn-group-toggle col-sm-4'>
                    <label class='btn btn-sm btn-outline-info'>
                        <input id='fixedQ' type='checkbox' data-toggle='q' autocomplete='off'> Fixed
                    </label>
                </div>
            </div>
            <div class='row'>
                <label class='col-sm-2' style='color:rgb(0,0,255)' id='qhklLabel'>Q</label>
                <input type='text' class='col-sm-3' id='qh'/>
                <input type='text' class='col-sm-3' id='qk'/>
                <input type='text' class='col-sm-3' id='ql'/>
            </div>
            <div class='row'>
                <label class='col-sm-2'>&Delta;E</label>
                <input type='text' class='col-sm-3' id='deltaE'/>
                <label class='col-sm-2'>meV</label>
                <div class='btn-group-toggle col-sm-4'>
                    <label class='btn btn-sm btn-outline-info'>
                        <input id='fixedDE' type='checkbox' data-toggle='de' autocomplete='off'> Fixed
                    </label>
                </div>
            </div>
            <div class='row'>
                <label class='col-sm-2'>Sense</label>
                <select class='col-sm-3' id='ssense'/>
                <label class='col-sm-2'/>
                <button class='col-sm-4 btn btn-sm btn-primary' id='pauseSpy' type='button'>Pause</button>
            </div>
        </div>`;

        // Using lambda function does not work here.
        myLayout.registerComponent('Scattering Parameters', function(container, state) {
            container.getElement().html(html);
        });

        viewManager.addView(this);
    }

    init() {

        // Set the colors. They are the same as in colors.js with darkness adjustement.
        // https://www.toutes-les-couleurs.com/couleurs-complementaires.php
        $("#kiLabel").css({'color' : "#7A39C4"});
        $("#kfLabel").css({'color' : "#9E9E2C"});
        $("#qLabel").css({'color' : "#2EABAB"});
        $("#qhklLabel").css({'color' : "#2EABAB"});

        // Update Q coordinate action
        $("#qh, #qk, #ql").keypress((event) => {
            if (event.which == 13 || event.keyCode == 13 ) {
                
                let qh = parseFloat($('#qh').val());
                let qk = parseFloat($('#qk').val());
                let ql = parseFloat($('#ql').val());
                let q = new THREE.Vector3(qh, qk, ql).applyMatrix3(matrix.bMatrix);
                
                this.updateQKiKf(scattering.ki, scattering.kf, qh, qk, ql, q.length(), scattering.ss);
            }
        });

        // Update Kf action
        $("#kf").keypress((event) => {
            if (event.which == 13 || event.keyCode == 13) {

                let kf = parseFloat($('#kf').val());

                this.updateQKiKf(scattering.ki, kf, scattering.qh, scattering.qk, scattering.ql, scattering.q, scattering.ss);
            }
        });

        // Update Ki action
        $("#ki").keypress((event) => {
            if (event.which == 13 || event.keyCode == 13) {

                let ki = parseFloat($('#ki').val());

                this.updateQKiKf(ki, scattering.kf, scattering.qh, scattering.qk, scattering.ql, scattering.q, scattering.ss);
            }
        });

        // Update Q action
        $("#qNorm").keypress((event) => {
            if (event.which == 13 || event.keyCode == 13) {

                let qnorm = parseFloat($('#qNorm').val());

                this.updateQKiKf(scattering.ki, scattering.kf, scattering.qh, scattering.qk, scattering.ql, qnorm, scattering.ss);
            }
        });

        // Update dE action
        $("#deltaE").keypress((event) => {
            if (event.which == 13 || event.keyCode == 13) {

                let ki = scattering.ki;    
                let kf = scattering.kf;
                let de = parseFloat($('#deltaE').val());

                // Calculate kf or kf depending on the fixed values.
                if (scattering.fixedKi) {
                    let kf2 = ki * ki - de / 2.072;
                    if (kf2 < 0) {
                        kf2 = 0;
                    }
                    kf = Math.sqrt(kf2);
                }
                else {
                    let ki2 = kf * kf + de / 2.072;
                    ki = Math.sqrt(ki2);
                }

                this.updateQKiKf(ki, kf, scattering.qh, scattering.qk, scattering.ql, scattering.q, scattering.ss);
            }
        });

        $("#ssense").append($("<option/>").val("1").html("1"));
        $("#ssense").append($("<option/>").val("-1").html("-1"));

        // Update scattering sense action
        $("#ssense").change((event) => {
            let ssb = $('#ssense').val() === '1';
            let ss = (ssb ? 1 : -1);

            this.updateQKiKf(scattering.ki, scattering.kf, scattering.qh, scattering.qk, scattering.ql, scattering.q, ss);
        });
        
        // Fix ki or kf action.
        $("#fixedKf, #fixedKi, #fixedQ, #fixedDE").change((event) => {
            if (event.target.checked) {
                $(event.target.parentElement).addClass('active');
            } else {
                $(event.target.parentElement).removeClass('active');
            }

            let kikfq = $(event.target).attr('data-toggle');
            let toggle = event.target.checked;

            PubSub.publish('fix-scattering', { kikfq, toggle });

            if (kikfq === 'kf') {
                $("#fixedKi").prop('checked', !toggle);
                $("#fixedKi").parent().toggleClass('active', !toggle);
            }

            if (kikfq === 'ki') {
                $("#fixedKf").prop('checked', !toggle);
                $("#fixedKf").parent().toggleClass('active', !toggle);
            }

        }).parent().hide();

        // Pause button.
        this._paused = false;

        // Hide the button if not in spy mode.
        if (!mode.spyMode) {
            $("#pauseSpy").hide();
        }

        $("#pauseSpy").click(() => {
            // Change the text.
            if (this._paused) {
                $("#pauseSpy").text('Pause');
            }
            else {
                $("#pauseSpy").text('Restart');
            }
            // Invert the paused state.
            this._paused = !this._paused;

            // Pause the scattering.
            PubSub.publish('pause-scattering', this._paused);
        });


        this.setInputs({
            ki: scattering.ki, kf: scattering.kf, q: scattering.q,
            qh: scattering.qh, qk: scattering.qk, ql: scattering.ql,
            fixedKi: scattering.fixedKi, fixedKf: scattering.fixedKf, fixedQ: scattering.fixedQ, fixedDE: scattering.fixedDE,
            ss: scattering.ss
        });

    }
    
    /**
     * Updates the points with the new ki value.
     */
    updateQKiKf(ki, kf, qh, qk, ql, qnorm, ss) {

        // Get q in R frame.
        let q = new THREE.Vector3(qh, qk, ql).applyMatrix3(matrix.bMatrix);
        q.setLength(qnorm);

        // Calculate the new points.
        let result = calculateKiKf(q, ki, kf, ss, matrix.uMatrix_1, matrix.uMatrix);

        // Send new points.
        let kiPos = result.kiPoints.start;
        let qPos = result.kfPoints.end;
        let qVec = qPos.clone().applyMatrix3(matrix.bMatrix_1);

        PubSub.publish('update-panel', {
            qPoints: qVec, 
            kiPoints: {
                start: kiPos,
                end: new THREE.Vector3()
            },
            kfPoints: {
                start: kiPos,
                end: qPos
            },
            ss
        });
    }

    /**
     * Set all scattering parameters inputs.
     */
    setInputs(value) {
        $('#ki').val(value.ki.toFixed(3));
        $('#kf').val(value.kf.toFixed(3));
        $('#qNorm').val(value.q.toFixed(3));
        $('#qh').val(value.qh.toFixed(3));
        $('#qk').val(value.qk.toFixed(3));
        $('#ql').val(value.ql.toFixed(3)); 

        // Recalculate dE
        let dE = 2.072 * (value.ki * value.ki - value.kf * value.kf);
        $('#deltaE').val(dE.toFixed(3));

        $("#fixedKi").parent().toggleClass('active', value.fixedKi);
        $("#fixedKf").parent().toggleClass('active', value.fixedKf);
        $("#fixedQ").parent().toggleClass('active', value.fixedQ);
        $("#fixedDE").parent().toggleClass('active', value.fixedDE);

        $("#ssense").val(value.ss.toFixed(0));
    }

    /**
     * Get all scattering parameters inputs and store them into
     * an object.
     */
    getInputs() {
        return {
            ki: parseFloat($('#ki').val()),
            kf: parseFloat($('#kf').val()),
            q: parseFloat($('#qNorm').val()),
            qh: parseFloat($('#qh').val()),
            qk: parseFloat($('#qk').val()),
            ql: parseFloat($('#ql').val())
        };
    }

    setMode(mode) {
        
        if (mode === 'snap') {
            $("#fixedKf, #fixedKi, #fixedQ, #fixedDE").prop('checked', false);
            $("#fixedKf, #fixedKi, #fixedQ, #fixedDE").parent().removeClass('active').hide();
            PubSub.publish('fix-scattering', { kikf: 'ki', toggle: false });
            PubSub.publish('fix-scattering', { kikf: 'kf', toggle: false });
            PubSub.publish('fix-scattering', { kikf: 'q', toggle: false });
            PubSub.publish('fix-scattering', { kikf: 'de', toggle: false });
        }
        else {
            $("#fixedKf, #fixedKi, #fixedQ, #fixedDE").parent().show();
        }
    }

    render() {
    }
}

module.exports = ScatteringPanelView;