/*
 * Not used.
 */

const EPSILON = 1.0E-06;
//const CMP_EPSILON = 1.0E-08;

const TWO_PI = 2.0*Math.PI;
const PI_ON_2 = Math.PI/2.0;

const K2V     = 629.622368;        /* Convert k[1/AA] to v[m/s] */
const VS2E    = 5.22703725e-6;     /* Convert (v[m/s])**2 to E[meV] */
const DELTA_E_CONST = VS2E*K2V*K2V;/* with help from Emmanuel Fahri */


class TASFormulas {
    constructor() {
        this.x_Sample;
        this.y_Sample;

        this.x_Analyser_s;
        this.y_Analyser_s;
        this.x_Analyser;
        this.y_Analyser;

        this.x_Detector_a;
        this.y_Detector_a;
        this.x_Detector_s;
        this.y_Detector_s;
        this.x_Detector;
        this.y_Detector;
    }

    /**
     * @param a2
     *            A2 in radian
     * @param a4
     *            A4 in radian
     * @param a6
     *            A6 in radian
     * @param l24
     *            in m, length between A2 and A4 rotation axes
     * @param l46
     *            in m, length between A4 and A6 rotation axes
     * @param l6d
     *            in m, length between A6 rotation axis and detector tube
     */
    instrument_position(a2, a4, a6, l24, l46, l6d) {

        // Sample Position
        this.x_Sample = l24 * Math.cos(a2);
        this.y_Sample = l24 * Math.sin(a2);

        // Analyser Position
        // in Sample system:
        this.x_Analyser_s = l46 * Math.cos(a4);
        this.y_Analyser_s = l46 * Math.sin(a4);
        // in Monochromator system
        this.x_Analyser = this.x_Sample + this.x_Analyser_s * Math.cos(a2) - this.y_Analyser_s * Math.sin(a2);
        this.y_Analyser = this.y_Sample + this.x_Analyser_s * Math.sin(a2) + this.y_Analyser_s * Math.cos(a2);

        // Detector position
        // in Analyser system:
        this.x_Detector_a = l6d * Math.cos(a6);
        this.y_Detector_a = l6d * Math.sin(a6);
        // in Sample system
        this.x_Detector_s = this.x_Analyser_s + this.x_Detector_a * Math.cos(a4) - this.y_Detector_a * Math.sin(a4);
        this.y_Detector_s = this.y_Analyser_s + this.x_Detector_a * Math.sin(a4) + this.y_Detector_a * Math.cos(a4);
        // in Monochromator system:
        this.x_Detector = this.x_Sample + this.x_Detector_s * Math.cos(a2) - this.y_Detector_s * Math.sin(a2);
        this.y_Detector = this.y_Sample + this.x_Detector_s * Math.sin(a2) + this.y_Detector_s * Math.cos(a2);

    }
    // Formula 4.17
    // Formula 4.18
    static volume_of_unit_cell(a, b, c, alpha, beta, gamma) {
        let alpha_cos = Math.cos(alpha);
        let beta_cos = Math.cos(beta);
        let gamma_cos = Math.cos(gamma);
        let  volume = Math.abs(a) * Math.abs(b) * Math.abs(c) * Math.sqrt(1 - Math.pow(alpha_cos, 2.0)
                -Math.pow(beta_cos, 2.0) - Math.pow(gamma_cos, 2.0) + 2 * alpha_cos * beta_cos * gamma_cos);
        return volume;
    }
    // Formula 4.5 to 4.7, 4.11 to 4.13
    // Used with permutation on a, b, c
    static direct_from_to_rec_cell_lattice(d1, d2, angle, volume) {
        let d_rec=1.0;
        if (volume > 0.0) {
            d_rec = TWO_PI * d1 * d2 * Math.sin(angle) / volume;
        } else {
            /*Throw exception ?*/;
        }
        return d_rec;
    }

    /**
     * Formula 4.8 to 4.10, 4.14 to 4.16
     * Used with permutation on alpha, beta, gamma
     */
    static direct_from_to_rec_cell_angle(a1, a2, a3) {
        let a_rec;
        a_rec = Math.acos((Math.cos(a1) * Math.cos(a2) - Math.cos(a3)) / (Math.sin(a1) * Math.sin(a2)));
        return a_rec;
    }

    /**
     * Formula 4.4
     * Compute delta_E from ki_len and kf_len, in mEv
     *
     * @param ki_len	Ki length in angstrom-1
     * @param kf_len	Kf length in angstrom-1
     * @return delta_e	energy transfer (in meV)
     */
    static deltaEfromKilenKflen(ki_len, kf_len) {
        let delta_e = DELTA_E_CONST	* (ki_len * ki_len - kf_len * kf_len);
        return delta_e;
    }

    /**
     * Formula 4.4 bis
     * Compute ki_len from delta_E and kf_len, in angstrom-1
     *
     * @param delta_e	energy transfer (in meV)
     * @param kf_len	Kf length in angstrom-1
     * @return ki_len	Ki length in angstrom-1
     */
    static KilenFromDeltaEKflen(delta_E, kf_len) {
        let ki_len = 0.0;
        let arg_sqrt = kf_len * kf_len + delta_E / DELTA_E_CONST;

        if (arg_sqrt < 0) {
            console.log("KilenFromDeltaEKflen: parameter out of range");
        } else {
            ki_len = Math.sqrt(arg_sqrt);
        }
        return ki_len;
    }

    /**
     * Formula 4.4 ter
     * Compute kf_len from delta_E and ki_len, in angstrom-1
     *
     * @param delta_e	energy transfer (in meV)
     * @param ki_len	Ki length in angstrom-1
     * @return kf_len	Kf length in angstrom-1
     * @throws Exception
     */
    static KflenFromDeltaEKilen(delta_E, ki_len) {
        let kf_len = 0.0;
        let arg_sqrt = ki_len * ki_len - delta_E / DELTA_E_CONST;

        if (arg_sqrt < 0 ) {
            console.log("KflenFromDeltaEKilen: parameter out of range");
        } else {
            kf_len = Math.sqrt(arg_sqrt);
        }
        return kf_len;
    }


    /**
     * Formula 4.24, 4.29
     * Compute angle from k_len, in radian, <-PI,PI>
     *
     * @param k_len		Ki or Kf length in angstrom-1
     * @param S			Scattering sense
     * @param d			Lattice spacing
     * @return angle	Bragg angle
     */
    static anglefromKlen(k_len, S, d) {
        let angle=0.0;
        if 	( k_len * d < Math.PI ) {
            console.log("anglefromKlen: parameter out of range");
            angle = S * PI_ON_2;
        } else {
            angle =  S * Math.asin(Math.PI/(k_len * d));
        }
        return angle;
    }

    /**
     * Formula 4.31, 4.32
     * Compute ki_len or kl_len from angle, in angstrom-1
     *
     * @param angle		angle in radian
     * @param d			Lattice spacing
     * @return angle	Bragg angle
     */
    static KlenFromAngle(angle, d) {
        if 	( Math.abs(angle) < EPSILON || Math.abs(Math.PI-angle) < EPSILON ) {
            console.log("KlenFromAngle: parameter out of range");
            return  Number.POSITIVE_INFINITY;
        }
        k_len =  Math.PI / ( d * Math.sin(angle) );
        return Math.abs(k_len);// TODO check this abs()
    }



    /**
     * Formula 4.26
     * Compute A4 from ki norm, kf norm, Q norm , in radian
     *
     * @param ki_len	ki norm in angstrom-1
     * @param kf_len	kf norm in angstrom-1
     * @param q_len		q norm in angstrom-1
     * @param S			Scattering sense
     * @return a4		angle in radian
     */
    static A4FromKilKflQlS(ki_len, kf_len, q_len, S) {
        let a4;
        let arg_acos = (ki_len * ki_len + kf_len * kf_len - q_len * q_len) / (2.0 * ki_len * kf_len) ;
        if 	( Math.abs(arg_acos) > 1.0 ){
            console.log("A4FromKilKflQlS: parameter out of range");
        } else {
            a4 = S * Math.acos( arg_acos );
        }
        return a4;
    }

    /**
     * Formula 4.27
     * Compute A3 from ki norm, kf norm, q norm , in radian
     *
     * @param ki_len	ki norm in angstrom-1
     * @param kf_len	kf norm in angstrom-1
     * @param qx		q x coordinate in angstrom-1
     * @param qy		q y coordinate in angstrom-1
     * @param q_len		q norm in angstrom-1
     * @return a3		angle in radian
     */
    static A3FromKilKflQxQyQlSs(ki_len, kf_len, qx, qy, q_len, signum_a4) {
        let a3;
        let arg_acos =  (ki_len*ki_len + q_len*q_len - kf_len*kf_len)/( 2.0*ki_len*q_len);

        if 	( Math.abs(arg_acos) > 1) {
            console.log("A3FromKilKflQxQyQlSs: parameter out of range");
        } else {
            a3 = Math.PI - Math.acos( arg_acos ) * signum_a4 - Math.atan2(qy,qx);
        }

// Keep it between -PI and PI
        while (a3 > Math.PI) {
            a3 -= TWO_PI;
        }
        while (a3 < -Math.PI) {
            a3 += TWO_PI;
        }

        return a3;
    }

    /**
     * Formula 4.2x
     * Compute A3 from ki x and ki y
     *
     * @param ki_x		x coordinate of ki
     * @param ki_y		y coordinate of ki
     * @return a3		angle in radian
     */
    static A3FromKixKiy(ki_x, ki_y) {
        let a3;
        a3 = - Math.atan2(  ki_y,  ki_x);

        // minus on angle because we go from Ki to A
        return a3;
    }


    /**
     * Formula 4.33
     * Compute q norm from ki norm, kf norm and angle A4, in angstrom-1
     *
     * @param ki_len	ki norm in angstrom-1
     * @param kf_len	kf norm in angstrom-1
     * @param a4		angle
     * @return qn		Q norm
     */
    static QlFromKilKflA4(ki_len, kf_len, a4) {
        let q_len = Math.sqrt(ki_len * ki_len + kf_len * kf_len - 2.0 * ki_len * kf_len * Math.cos(a4) );
        return q_len;
    }

    /**
     * Formula 4.34, 4.35 (sub part)
     * Compute angle from ki norm, kf norm , q norm, a3, sign of a4 in radian
     *
     * @param ki_len
     * @param kf_len
     * @param q_len
     * @param a3
     * @param signum a4
     * @return
     */
    static QangleFromKilKflQlA3(ki_len, kf_len, q_len, a3, signum_a4) {
        let angle;
        let arg_acos =  - ((kf_len*kf_len - q_len*q_len - ki_len*ki_len) / (2.0*q_len*ki_len)) ;
        if 	( Math.abs(arg_acos) > 1) {
            console.log("QangleFromKilKflQlA3: parameter out of range");
            if (arg_acos>=0) {
                angle = - a3 ;
            } else {
                angle = - Math.PI*signum_a4 - a3;
            }
        } else {
            angle = - Math.acos( arg_acos )*signum_a4 - a3;
        }
        return angle;
    }


    /**
     * Formula 4.34, 4.35 variante
     * Compute Qabs from ki norm, kf norm , q norm, a3, a4 in radian
     *
     * @param ki_len
     * @param kf_len
     * @param q_len
     * @param a3
     * @param a4
     * @return
     */
    static QabsFromKilKflA3A4_Bis(ki_len, kf_len, a3, a4) {
        let new_q_abs;

        let qpar = ki_len - kf_len * Math.cos(a4);
        let qperp = - kf_len * Math.sin(a4);


        let  q_abs_x = -qpar*Math.cos(a3)-qperp*Math.sin(a3);
        let q_abs_y = qpar*Math.sin(a3)-qperp*Math.cos(a3);

        new_q_abs = new Vector2D(q_abs_x, q_abs_y);
        return new_q_abs;
    }




    /**
     * Formula 4.xx
     * Compute Ss from ki , kf
     *
     *
     *
     * @param kiX - x value of ki
     * @param kiY - y value of ki
     * @param kfX - x value of kf
     * @param kfY - y value of kf
     * @return Ss
     */
    static SsFromKiKf(kiX,kiY, kfX, kfY) {
        // By definition, Ss is positive if kf is at the left of ki, negative at right.
        // use sign of determinant between ki and q_abs
        let d = kiX * kfY - kiY * kfX;
        let Ss = d>=0.0 ? 1 : -1;// Note : Math.signum returns 0 if argument is 0; this can produce undesirable effect
        return Ss;
    }



    /**
     * 4.36
     * Compute a new Ki vector from its length and its angle
     * I don't know if this is in degrees or radians
     */
    static KiFromKilA3(ki_len, a3) {
        return new Vector2D(
            ki_len * Math.cos( - a3),
            ki_len * Math.sin( - a3));
    }


    /*
     * Getters
     *
     *
     *
     */
    static get CMP_EPSILON() {
        return CMP_EPSILON;
    }

    get getSource_position() {
        return new Vector2D(0.0, 0.0);
    }

    get getSample_position() {
        return new Vector2D(this.x_Sample, this.y_Sample);
    }

    get getAnalyser_position() {
        return new Vector2D(this.x_Analyser, this.y_Analyser);
    }

    get getDetector_position() {
        return new Vector2D(this.x_Detector, this.y_Detector);
    }

    get getUB_matrix() {
        return this.UB_matrix;
    }
}

module.exports = TASFormulas;